import contacto.helpers as hlp
from helpers import fixture
import pytest


def test_refspec_parse():
    with pytest.raises(Exception):
        hlp.parse_refspec('///')
        hlp.parse_refspec('/Group/Entity/Attribute')
    assert hlp.parse_refspec('Group') == ('Group', None, None)


def test_ref_parse():
    assert hlp.parse_ref('Group/Entity')[0] is hlp.DType.EXREF
    assert hlp.parse_ref('Group/Entity/Attribute')[0] is hlp.DType.AXREF
    with pytest.raises(Exception):
        hlp.parse_ref('Group')
        hlp.parse_ref('///')


def test_valspec_parse():
    fix = fixture('tinybin.dat')
    t, d =  hlp.parse_valspec(f"FILE:{fix}")
    assert t == hlp.DType.BIN and d == fix.read_bytes()
    t, d = hlp.parse_valspec(f"URL:file:{fix}")
    assert t == hlp.DType.BIN and d == fix.read_bytes()
    t, d = hlp.parse_valspec("REF:G/E/A")
    assert t == hlp.DType.AXREF and d == ('G', 'E', 'A')
    with pytest.raises(Exception):
        hlp.parse_valspec("REF:Group/Entity//")
        hlp.parse_valspec("REF:Group")



from helpers import mkstor, yml_fixture
from contacto.serial import Serial

def test_import_export():
    stor = mkstor()
    ser = Serial(stor)
    with open(yml_fixture('test'), 'r') as f:
        assert ser.import_yaml(f)
    assert stor.get_attribute('Family', 'Dad', 'age')

    with open(yml_fixture('temp'), 'w') as f:
        assert ser.export_yaml(f)
    
    # reimport our export
    stor2 = mkstor()
    ser2 = Serial(stor2)
    with open(yml_fixture('temp'), 'r') as f:
        assert ser2.import_yaml(f)
    
    for gname in stor.groups.keys():
        assert gname in stor2.groups



import contacto.storage as storage
from contacto.helpers import DType
from helpers import mkstor, fixture
import pytest


@pytest.fixture
def stor():
    return mkstor()


def __test_storage_elem(stor, elem):
    assert elem.get_storage() is stor
    assert elem.get_conn() is stor.db_conn
    s, elem.name = elem.name, ''
    elem.read()
    assert s == elem.name


def test_group(stor):
    grp_name = 'MYGROUP'
    assert not stor.get_group(grp_name)
    grp = stor.create_group_safe(grp_name)
    assert grp and grp is stor.get_group(grp_name)
    assert grp_name in stor.groups
    assert str(grp) == grp_name
    __test_storage_elem(stor, grp)
    
    grp2_name = 'MYGROUP2'
    grp2 = stor.create_group_safe(grp2_name)
    assert stor.get_group(grp2_name)
    assert grp.merge_safe(grp2)
    assert not stor.get_group(grp2_name)


def test_entity(stor):
    gname, ename = 'MYGROUP', 'MYENTITY'
    grp = stor.create_group_safe(gname)
    ent = grp.create_entity_safe(ename)
    assert ent and ent is stor.get_entity(gname, ename)
    assert ename in grp.entities
    __test_storage_elem(stor, ent)
    
    thumb = fixture('cat.jpg').read_bytes()
    ent.thumbnail = thumb
    ent.update()

    gname2 = 'MYGROUP2'
    grp2 = stor.create_group_safe(gname2)
    ent2 = grp2.create_entity_safe(ename)
    assert grp2.merge_safe(grp)

    assert ent2.thumbnail == thumb


def test_attribute(stor):
    gn, en = 'MYGROUP', 'MYENTITY'
    adata = 'MYATTR', DType.TEXT, 'MYVALUE'
    grp = stor.create_group_safe(gn)
    ent = grp.create_entity_safe(en)
    att = ent.create_attribute_safe(*adata)
    assert att and att is stor.get_attribute(gn, en, att.name)
    __test_storage_elem(stor, att)

    # thumb test
    thumb = fixture('cat.jpg').read_bytes()
    adata2 = 'thumbnail', DType.BIN, thumb
    att2 = ent.create_attribute_safe(*adata2)
    assert att2
    assert ent.thumbnail == thumb

    # AXREF test
    adata3 = 'areference', DType.AXREF, att2
    att3 = ent.create_attribute_safe(*adata3)
    assert att3
    assert att2 is stor.elem_from_refid(att3.type, att3.data.id)

    # EXREF test
    adata4 = 'ereference', DType.EXREF, ent
    att4 = ent.create_attribute_safe(*adata4)
    assert att4
    assert ent is stor.elem_from_refid(att4.type, att4.data.id)

    # cascade thumb delete test
    assert att2.delete_safe()
    assert not stor.get_attribute(gn, en, att2.name)
    assert not stor.get_attribute(gn, en, att3.name)
    assert not ent.thumbnail

    # EXREF merge tests
    assert att.merge_safe(att4)
    assert att.type is DType.EXREF

    en2 = 'MYENTITY2'
    ent2 = grp.create_entity_safe(en2)
    assert ent2.merge_safe(ent)
    att5 = stor.get_attribute(gn, en2, att.name)
    assert att5 and att5.data == ent2


def test_storage(stor):
    names = [str(i) for i in range(50)]
    for name in names:
        assert stor.create_group_safe(name)
    stor.groups = {}
    stor.reload()
    for name in names:
        assert stor.get_group(name)



"""Serialization features: import, export, human-readable dumping.
"""

import yaml
import click
import os
from urllib.request import urlopen
from .helpers import DType, Scope, parse_valspec, attr_val_str, print_error


class Serial:
    """Serialization handler, accepts a data storage.
    """
    def __init__(self, storage):
        """Constructor, save provided storage
        """
        self.storage = storage

    def export_yaml(self, file, max_scope=Scope.ATTRIBUTE, max_bin_size=0):
        """Exports storage in YAML format into a file.

        Maximum scope may be provided to cut away attributes or even entities.
        If maximum binary size is set, binary data bigger than this limit will
        be dumped into files in the dump file's directory and linked by refs.

        :param file: file to dump to
        :type  file: class:`io.TextIOWrapper`
        :param max_scope: rightmost tree scope to export
        :type  max_scope: class:`contacto.helpers.Scope`, optional
        :param max_bin_size: maximum binary data size to inline
        :type  max_bin_size: int, optional
        :return: success
        :rtype:  bool
        """
        data = {}
        try:
            # group scope
            for gname, group in self.storage.groups.items():
                d_group = {}
                data[gname] = d_group
                # entity scope
                if max_scope <= Scope.GROUP:
                    continue
                for ename, entity in group.entities.items():
                    d_entity = {}
                    d_group[ename] = d_entity
                    # attribute scope
                    if max_scope <= Scope.ENTITY:
                        continue
                    for aname, attribute in entity.attributes.items():
                        value = attribute.data
                        if attribute.type.is_xref():
                            value = f"REF:{value}"
                        elif (attribute.type is DType.BIN
                              and max_bin_size > 0
                              and len(value) > max_bin_size):
                            # dump data to file and reference it
                            dirp = os.path.dirname(os.path.realpath(file.name))
                            fname = os.path.join(dirp, f"{attribute.id}.dat")
                            with open(fname, 'wb') as f:
                                f.write(value)
                            value = f"FILE:{fname}"
                        d_entity[aname] = value

            yaml.safe_dump(data, file)
        except Exception as e:
            print_error(e)
            return False
        return True

    def dump(self, direct=False, lscope=Scope.GROUP, rscope=Scope.ATTRIBUTE):
        """Dumps storage in a human-readable form to stdout.

        This dump type may be scoped from the left if a specific element
        is directly requested (lscope), such as with a G/E refspec.

        Right scope (rscope) cuts data from the right (starting with attrs.).
        It is possible to dump entities only (<E, E>), entities and attributes
        (<E, A>) or just attribute values (<AV, A>)...

        Passing the direct flag causes attributes to be printed without names.
        Desirable when printing a single-scoped attr's value (<AV, A>).

        :param direct: do not print attribute names
        :type  direct: bool, optional
        :param lscope: leftmost tree scope to export
        :type  lscope: class:`contacto.helpers.Scope`, optional
        :param rscope: rightmost tree scope to export
        :type  rscope: class:`contacto.helpers.Scope`, optional
        """
        if rscope < Scope.ATTRIBUTE and lscope > rscope:
            lscope = rscope
        # don't output directly unless a direct scope is used
        direct &= lscope is Scope.ATTR_VAL

        scope = 0

        def prefix():
            return '' if scope == 0 else f"{(scope - 1) * '  '}- "

        for gname, group in sorted(self.storage.groups.items()):
            if lscope <= Scope.GROUP:
                click.echo(f"{prefix()}{gname}")
                if rscope == Scope.GROUP:
                    continue
                else:
                    scope += 1
            for ename, entity in sorted(group.entities.items()):
                if lscope <= Scope.ENTITY:
                    click.echo(f"{prefix()}{ename}")
                    if rscope == Scope.ENTITY:
                        continue
                    else:
                        scope += 1
                max_len = 0
                for aname in entity.attributes.keys():
                    max_len = max(max_len, len(aname))
                for aname, attr in sorted(entity.attributes.items()):
                    val = attr_val_str(attr, direct)
                    if lscope <= Scope.ATTRIBUTE:
                        pad = (max_len - len(aname)) * ' '
                        click.echo(f"{prefix()}{aname}{pad}: {val}")
                    else:
                        click.echo(val)
                if lscope <= Scope.ENTITY:
                    scope -= 1
            if lscope <= Scope.GROUP:
                scope -= 1

    def import_yaml(self, file):
        """Imports YAML data from a file into the storage.

        :param file: YAML file
        :type  file: class:`io.TextIOWrapper`
        :return: success
        :rtype:  bool
        """
        data = None
        try:
            data = yaml.safe_load(file)
        except Exception as e:
            print_error(e)
            return False

        try:
            with self.storage.db_conn:
                self.__import_yamldata(data)
        except Exception as e:
            # import error, reload in-memory data
            self.storage.reload()
            print_error(e)
            return False
        return True

    def __import_yamldata(self, data):
        """Imports extracted (and parsed) YAML data.
        """
        xref_queue = []
        for gname, d_group in data.items():
            group = self.storage.get_group(gname) or \
                    self.storage.create_group(gname)
            if d_group is None:
                continue
            for ename, d_entity in d_group.items():
                entity = self.storage.get_entity(gname, ename) or \
                         group.create_entity(ename)
                if d_entity is None:
                    continue
                for aname, d_attr in d_entity.items():
                    atype, adata = self.__parse_yaml_attr(d_attr)
                    # XREFs may not resolve yet, bypass saving them for now
                    isXREF = atype.is_xref()
                    if isXREF:
                        ref_meta = atype, adata
                        atype, adata = DType.TEXT, '<XREF>'  # bypass XREFs
                    attr = self.storage.get_attribute(gname, ename, aname)
                    if attr is None:
                        attr = entity.create_attribute(aname, atype, adata)
                    else:
                        attr.name, attr.type, attr.data = aname, atype, adata
                        attr.update()
                    # queue XREFs for proper processing
                    if isXREF:
                        xref_queue.append((*ref_meta, attr))
        # process XREFs
        for ref_type, ref_data, attr in xref_queue:
            attr.type = ref_type
            attr.data = self.storage.get_from_rspec(ref_data)
            attr.update()

    def __parse_yaml_attr(self, d_attr):
        """Infers attribute value and type from YAML-parsed data.
        """
        if type(d_attr) is bytes:
            return DType.BIN, d_attr
        if type(d_attr) is str:
            return parse_valspec(d_attr)
        return DType.TEXT, str(d_attr)



"""Module used for searching within the storage tree.

It creates induced subgraphs from the tree by filtering out elements.
These subgraphs may then be viewed as search results.
"""
import copy
from .helpers import DType, fmatch
from .storage import Group, Entity, Attribute


class View:
    """A read-only representation of a Storage slice.

    Provides a "groups" member that mimics that of the Storage object.
    Therefore the View itself is a read-only storage root and may be passed
    to contexts that do not modify the tree, such as serialization.

    To filter the tree, set your filters first using the provided methods.
    """

    def __init__(self, storage):
        """Initialize with a storage containing the full tree
        """
        self.source = storage.groups
        self.reset()

    def reset(self):
        """Resets the View back to the full tree, discarding any prior scoping.

        Also discards any set filters.
        """
        self.index_filters = [None, None, None]
        self.value_predicates = [None, None, None]
        self.groups = self.source

    def empty(self):
        """Tests if no filters have been set

        :return: True if no filters are set
        :rtype:  bool
        """
        return self.index_filters == [None, None, None] and \
            self.value_predicates == [None, None, None]

    def set_index_filters(self, filters):
        """Sets string needles that match exactly on an element's name.

        :param filters: parsed generic refspec (G/E/A)
        :type filters:  tuple
        """
        self.index_filters = [
            new_f or old_f
            for new_f, old_f in zip(filters, self.index_filters)
        ]

    def set_name_filters(self, filters):
        """Sets string needles that match fuzzily on an element's name.

        This means a case-insensitive substring search.
        Example: "abc" matches an element named "zAbcDeF"

        :param filters: parsed generic refspec (G/E/A)
        :type filters:  tuple
        """
        def sel(filt):
            if filt:
                return lambda x: fmatch(filt, x.name)
            return None

        preds = [sel(filt) for filt in filters]
        self.set_value_predicates(preds)

    def set_attr_value_filter(self, needle, fuzzy):
        """Sets a string needle that matches on a TEXT Attribute's value.

        Non-TEXT Attributes are not considered a match.
        May be specified both as exact and fuzzy search.
        An empty string is not considered a valid value filter.

        :param needle: search needle
        :type needle:  str
        :param fuzzy: use fuzzy search for the needle
        :type fuzzy:  bool
        """
        if not needle:
            return

        def pred(x):
            dtype, val = x.get()
            if dtype is not DType.TEXT:
                return False
            if fuzzy:
                return fmatch(needle, val)
            return needle == val

        self.set_value_predicates((None, None, pred))

    def set_value_predicates(self, preds):
        """Sets generic predicates that determine if an element is matched.

        Receives 3 predicates, one for each tree element type.
        This method may be called multiple times to stack multiple predicates
        on a single element. Stacked predicates are joined with logical AND.

        Example predicates:

        `L(attr): attr is binary`

        `L(entity): entity has thumbnail`

        :param preds: element predicates
        :type  preds: tuple
        """
        def join(new_p, old_p):
            if old_p and new_p:
                return lambda x: old_p(x) and new_p(x)
            return new_p or old_p

        self.value_predicates = [
            join(new_p, old_p)
            for new_p, old_p in zip(preds, self.value_predicates)
        ]

    def filter(self):
        """Generates an induced subgraph from the current tree
        using the saved filters
        """
        if self.empty():
            return
        ind_g, ind_e, ind_a = self.index_filters  # strings
        val_g, val_e, val_a = self.value_predicates  # predicates

        grps = {}
        iter_groups = self.groups
        # group index filter
        if ind_g:
            if ind_g not in iter_groups:
                self.groups = {}
                return
            iter_groups = {ind_g: iter_groups[ind_g]}
        for gname, group in iter_groups.items():
            # group value filter
            if val_g and not val_g(group):
                continue
            iter_entities = group.entities
            # entity index filter
            if ind_e:
                if ind_e not in iter_entities:
                    continue
                iter_entities = {ind_e: iter_entities[ind_e]}
            ents = {}
            for ename, entity in iter_entities.items():
                # entity value filter
                if val_e and not val_e(entity):
                    continue
                iter_attributes = entity.attributes
                # attribute index filter
                if ind_a:
                    if ind_a not in iter_attributes:
                        continue
                    iter_attributes = {ind_a: iter_attributes[ind_a]}
                attrs = {}
                for aname, attr in iter_attributes.items():
                    # attribute value filter
                    if val_a and not val_a(attr):
                        continue
                    attrs[aname] = copy.copy(attr)
                nofilt = len(iter_attributes) == 0 and not val_a
                if len(attrs) > 0 or nofilt:
                    ent = copy.copy(entity)
                    ent.attributes = attrs
                    for attr in attrs.values():
                        attr.parent = ent
                    ents[ename] = ent
            nofilt = len(iter_entities) == 0 and not (val_e or ind_a or val_a)
            if len(ents) > 0 or nofilt:
                grp = copy.copy(group)
                grp.entities = ents
                for ent in ents.values():
                    ent.parent = grp
                grps[gname] = grp
        self.groups = grps



"""A Qt5 frontend for Contacto

For now it it read-only.
"""
from PyQt5 import QtWidgets, uic
import sys
import pkgutil
import io
from .storage import Storage
from .serial import Serial
from .helpers import DType, attr_val_str, print_error


MAIN_UI = 'resources/mainwindow.ui'
ABOUT_HTML = 'resources/about.html'


class GUI:
    """Wrapper for all GUI logic
    """
    def __init__(self):
        """GUI initializer

        Allows passing DB name from args
        """
        self.app = QtWidgets.QApplication(sys.argv)
        self.app.setApplicationName("Contacto")

        # initialize from args
        if len(sys.argv) > 1:
            fname = sys.argv[1]
        else:
            fname = ':memory:'
        self.open(fname)

        script = pkgutil.get_data(__name__, MAIN_UI).decode('utf-8')
        sio = io.StringIO(script)

        # load main window
        self.window = QtWidgets.QMainWindow()
        uic.loadUi(sio, self.window)

        # load main UI areas
        self.tree = self.window.findChild(QtWidgets.QTreeWidget, 'treeWidget')
        self.tree.setColumnWidth(0, 200)
        self.init_tree()

        # actions
        act = self.window.findChild(QtWidgets.QAction, 'actionOpen')
        act.triggered.connect(self.action_open)
        act = self.window.findChild(QtWidgets.QAction, 'actionAbout')
        act.triggered.connect(self.action_about)
        act = self.window.findChild(QtWidgets.QAction, 'actionImport_from')
        act.triggered.connect(self.action_import)
        act = self.window.findChild(QtWidgets.QAction, 'actionExport_to')
        act.triggered.connect(self.action_export)

    @staticmethod
    def __type_to_str(dtype):
        """Returns a string representation of attribute type

        :param dtype: attribute data type
        :type dtype:  class:`contacto.helpers.DType`
        :return: string representation
        :rtype:  str
        """
        tstr = '???'
        if dtype is DType.BIN:
            tstr = "binary"
        elif dtype is DType.TEXT:
            tstr = "text"
        elif dtype is DType.AXREF:
            tstr = "-> attr"
        elif dtype is DType.EXREF:
            tstr = "-> entity"
        return f"Attr ({tstr})"

    def open(self, dbname):
        """Open a new Storage around a filename

        :param dbname: name of database holding raw data
        :type  dbname: str
        """
        self.storage = Storage(dbname)
        self.serial = Serial(self.storage)
        self.filename = dbname

    def init_tree(self):
        """Loads contact data into the tree widget
        """
        for gname, group in sorted(self.storage.groups.items()):
            gitem = QtWidgets.QTreeWidgetItem(self.tree)
            gitem.setText(0, gname)
            gitem.setText(1, "Group")
            for ename, entity in sorted(group.entities.items()):
                eitem = QtWidgets.QTreeWidgetItem(gitem)
                eitem.setText(0, ename)
                eitem.setText(1, "Entity")
                for aname, attr in sorted(entity.attributes.items()):
                    aitem = QtWidgets.QTreeWidgetItem(eitem)
                    aitem.setText(0, aname)
                    aitem.setText(1, self.__type_to_str(attr.type))
                    aitem.setText(2, attr_val_str(attr, False))

    def action_open(self):
        """Qt action: open new DB
        """
        fname = QtWidgets.QFileDialog.getOpenFileName()[0]
        if not fname:
            return
        try:
            self.open(fname)
            self.init_tree()
        except Exception as e:
            print_error(str(e))
            QtWidgets.QMessageBox.critical(self.window, 'Load error', str(e))

    def action_about(self):
        """Qt action: display About dialog
        """
        about_str = pkgutil.get_data(__name__, ABOUT_HTML).decode('utf-8')
        QtWidgets.QMessageBox.about(self.window, 'About Contacto', about_str)

    def action_import(self):
        """Qt action: import a YAML file
        """
        yml_filt = "YAML files (*.yml)"
        fname = QtWidgets.QFileDialog.getOpenFileName(filter=yml_filt)[0]
        if not fname:
            return
        try:
            serial = Serial(self.storage)
            with open(fname, 'r') as f:
                if not serial.import_yaml(f):
                    raise Exception("YAML import failed")
        except Exception as e:
            QtWidgets.QMessageBox.critical(self.window, 'Import error', str(e))

    def action_export(self):
        """Qt action: export DB into a YAML file
        """
        yml_filt = "YAML files (*.yml)"
        fname = QtWidgets.QFileDialog.getSaveFileName(filter=yml_filt)[0]
        if not fname:
            return
        try:
            serial = Serial(self.storage)
            with open(fname, 'w') as f:
                if not serial.export_yaml(f):
                    raise Exception("YAML export failed")
        except Exception as e:
            QtWidgets.QMessageBox.critical(self.window, 'Export error', str(e))

    def run(self):
        """Object entry-point (open and run)
        """
        self.window.show()
        return self.app.exec()


def main():
    """GUI entry-point
    """
    gui = GUI()
    gui.run()


# allow running GUI directly via "python -m contacto.gui"
if __name__ == '__main__':
    main()



"""CLI interface using Click.
"""

import click
import sys
from .storage import Storage
from .helpers import parse_refspec, parse_valspec
from .helpers import DType, Scope, dump_lscope, refspec_scope
from .helpers import print_warning, print_error, get_plugins, run_plugins
from .view import View
from .serial import Serial


def group_set(storage, gname):
    """Creates or update a group.

    :param storage: used storage
    :type  storage: class:`contacto.storage.Storage`
    :param gname: group name
    :type  gname: str
    :return: created group
    :rtype:  class:`contacto.storage.Group`
    """
    return storage.get_group(gname) or \
        storage.create_group_safe(gname) or sys.exit(1)


def entity_set(storage, gname, ename, recursive):
    """Creates or update an entity.

    :param storage: used storage
    :type  storage: class:`contacto.storage.Storage`
    :param gname: group name
    :type  gname: str
    :param ename: entity name
    :type  ename: str
    :param recursive: create group if non-existent
    :type  recursive: bool
    :return: created entity
    :rtype:  class:`contacto.storage.Entity`
    """
    ent = storage.get_entity(gname, ename)
    if ent:
        return ent
    if recursive:
        grp = group_set(storage, gname)
    else:
        grp = storage.get_group(gname)
        if not grp:
            print_error(f"Group {gname} does not exist.")
            sys.exit(1)
    return grp.create_entity_safe(ename) or sys.exit(1)


def validate_refspec(ctx, param, value):
    """Validates a generic refspec (without restrictions).

    :param ctx: Click context
    :type  ctx: class:`click.Context`
    :param param: Click params
    :type  param: dict
    :param value: refspec
    :type  value: str
    :return: parsed refspec
    :rtype:  list
    """
    try:
        return parse_refspec(value)
    except Exception as e:
        raise click.BadParameter(str(e))


def validate_full_refspec(ctx, param, value):
    """Validates a full refspec (specified from the left).

    :param ctx: Click context
    :type  ctx: class:`click.Context`
    :param param: Click params
    :type  param: dict
    :param value: refspec
    :type  value: str
    :return: parsed refspec
    :rtype:  list
    """
    p_rspec = validate_refspec(ctx, param, value)
    if refspec_scope(p_rspec):
        return p_rspec
    raise click.BadParameter('Fully-specified refspec required')


@click.group()
@click.option('-o', '--open', 'dbname', type=click.Path(exists=False),
              required=True, help='Path to storage file')
@click.pass_context
def main_cmd(ctx, dbname):
    """Contacto CLI: manage your contacts in the console."""

    ctx.ensure_object(dict)
    ctx.obj['storage'] = Storage(dbname)


@main_cmd.command(name='get')
@click.option('-s', '--scope', help='Desired output scope.',
              type=click.Choice(['attr', 'ent', 'grp']),
              default='attr', show_default=True)
@click.option('-f', '--fuzzy', callback=validate_refspec,
              help='Refspec for fuzzy element matching.')
@click.option('-v', '--value', help='Match attributes by value TEXT.')
@click.option('-V', 'val_fuzzy', is_flag=True,
              help='Fuzzy match attributes by value (use with -v).')
@click.option('-r', '--raw', is_flag=True,
              help='Dump raw data. Full REFSPEC must be given.')
@click.option('-y', '--yaml', help='Output data in YAML format.', is_flag=True)
@click.argument('refspec', callback=validate_refspec, required=False)
@click.pass_context
def get_cmd(ctx, scope, fuzzy, value, val_fuzzy, raw, yaml, refspec):
    """Fetch and print matching elements"""

    scope = Scope.from_str(scope)
    if val_fuzzy and not value:
        print_warning('-V must be used with -v')

    view = View(ctx.obj['storage'])
    view.set_index_filters(refspec)
    view.set_name_filters(fuzzy)
    view.set_attr_value_filter(value, val_fuzzy)
    view.filter()

    serial = Serial(view)
    if yaml:
        serial.export_yaml(sys.stdout, max_scope=scope)
    else:
        serial.dump(raw, dump_lscope(refspec), scope)


@main_cmd.command(name='set')
@click.option('-r', '--recursive',
              help='Create elements recursively.', is_flag=True)
@click.option('-b', '--binary',
              help='Read binary data (use with -i).', is_flag=True)
@click.option('-i', '--stdin', help='Read VALUE from stdin.', is_flag=True)
@click.option('-R', '--rotate', help='Rotate attribute value.', is_flag=True)
@click.argument('refspec', callback=validate_full_refspec)
@click.argument('value', required=False)
@click.pass_context
def set_cmd(ctx, recursive, binary, stdin, rotate, refspec, value):
    """Create or update a REFSPEC-specified element.

    VALUE sets thumbnails of entities and values of attributes."""

    if binary and not stdin:
        print_warning('Warning: -b must be used with -i')

    stor = ctx.obj['storage']
    if not stdin:
        try:
            vtype, vdata = parse_valspec(value)
        except Exception as e:
            print_error(e)
            sys.exit(1)
        if vtype and vtype.is_xref():
            vdata = stor.get_from_rspec(vdata)
            if not vdata:
                print_error("Invalid reference")
                sys.exit(1)
    elif not binary:
        vtype, vdata = DType.TEXT, sys.stdin.read()
    else:
        vtype, vdata = DType.BIN, sys.stdin.buffer.read()

    scope = refspec_scope(refspec)
    if scope == Scope.GROUP:
        group_set(stor, refspec[0])
    if scope == Scope.ENTITY:
        entity_set(stor, *refspec[:2], recursive)
    if scope == Scope.ATTRIBUTE:
        if not vtype:
            print_error(f'Attributes require supplied value.')
            sys.exit(1)
        attr = stor.get_attribute(*refspec)
        if attr:
            if vtype == attr.type or vdata == attr.data:
                return
            if rotate:
                attr.rotate_safe() or sys.exit(1)
            attr.type, attr.data = vtype, vdata
            attr.update_safe() or sys.exit(1)
            return
        entp = refspec[:2]
        if recursive:
            ent = entity_set(stor, *entp, recursive)
        else:
            ent = stor.get_entity(*entp)
            if not ent:
                print_error(f"Entity {'/'.join(entp)} does not exist.")
                sys.exit(1)
        ent.create_attribute_safe(refspec[2], vtype, vdata) or sys.exit(1)


@main_cmd.command(name='del')
@click.argument('refspec', callback=validate_full_refspec)
@click.pass_context
def del_cmd(ctx, refspec):
    """Delete a REFSPEC-specified element."""

    stor = ctx.obj['storage']
    elem = stor.get_from_rspec(refspec) or sys.exit(1)
    elem.delete_safe() or sys.exit(1)


@main_cmd.command(name='merge')
@click.argument('refspec_src', callback=validate_full_refspec)
@click.argument('refspec_dst', callback=validate_full_refspec)
@click.pass_context
def merge_cmd(ctx, refspec_src, refspec_dst):
    """Merge entity/group specified by REFSPEC_SRC into REFSPEC_DST."""

    stor = ctx.obj['storage']
    src = stor.get_from_rspec(refspec_src) or sys.exit(1)
    dst = stor.get_from_rspec(refspec_dst) or sys.exit(1)

    if type(src) != type(dst) or not hasattr(src, 'merge'):
        print_error('You can only merge 2 entities or 2 groups.')
        sys.exit(1)
    dst.merge_safe(src) or sys.exit(1)


@main_cmd.command(name='import')
@click.argument('file', type=click.File('r'), required=False)
@click.pass_context
def import_cmd(ctx, file):
    """Import YAML data from FILE or stdin."""

    serial = Serial(ctx.obj['storage'])
    serial.import_yaml(file or sys.stdin) or sys.exit(1)


@main_cmd.command(name='export')
@click.argument('file', type=click.File('w'), required=False)
@click.pass_context
def export_cmd(ctx, file):
    """Export YAML data to FILE or stdout. Similar to 'get -y'"""

    serial = Serial(ctx.obj['storage'])
    serial.export_yaml(file or sys.stdout) or sys.exit(1)


@main_cmd.command(name='plugin')
@click.option('-l', '--list', help='List available plugins', is_flag=True)
@click.argument('whitelist', nargs=-1)
@click.pass_context
def plugin_cmd(ctx, list, whitelist):
    """Run Contacto plugins, optionally specify which by name"""

    if list:
        click.echo(', '.join(get_plugins().keys()))
        return

    ok, nok = run_plugins(ctx.obj['storage'], whitelist)
    s_ok = '' if len(ok) == 0 else f" ({', '.join(ok)})"
    s_nok = '' if len(nok) == 0 else f" ({', '.join(nok)})"

    click.echo(f"Plugin summary: {len(ok)} successful{s_ok}, \
               {len(nok)} failed{s_nok}.")
    if len(nok) > 0:
        sys.exit(1)


def main():
    """CLI entrypoint, initializes the Click main command"""

    main_cmd(prog_name='contacto')



from helpers import mkstor, yml_fixture
from contacto.serial import Serial
from contacto.view import View
import pytest


@pytest.fixture
def view():
    stor = mkstor()
    ser = Serial(stor)
    with open(yml_fixture('test'), 'r') as f:
        assert ser.import_yaml(f)
    return View(stor)


def test_index_filters(view):
    assert view.empty()

    ifilts = 'Family', None, 'web'
    view.set_index_filters(ifilts)
    assert not view.empty()

    view.filter()
    assert 'Family' in view.groups
    assert 'Friends' not in view.groups
    grp = view.groups['Family']
    assert len(grp.entities) == 2
    for ent in grp.entities.values():
        assert 'web' in ent.attributes
        assert len(ent.attributes) == 1


def test_name_filters(view):
    nfilts = 'f', None, 'elationship'
    view.set_name_filters(nfilts)
    assert not view.empty()

    view.filter()
    assert len(view.groups) == 2 and 'Family' in view.groups
    fam = view.groups['Family']
    assert 'Mom' not in fam.entities and 'Dad' in fam.entities


def test_value_filters(view):
    view.set_attr_value_filter('Logan', True)
    assert not view.empty()

    view.filter()
    assert 'Family' in view.groups and len(view.groups) == 1
    grp = view.groups['Family']
    assert len(grp.entities) == 2

    view.reset()
    assert view.empty()
    view.set_attr_value_filter('45', False)
    view.filter()
    assert 'Family' in view.groups


def test_all(view):
    view.set_index_filters((None, None, 'relationship'))
    view.set_name_filters((None, 'B', None))
    view.set_attr_value_filter('ial', True)
    view.filter()
    assert 'Friends' in view.groups and len(view.groups) == 1
    assert 'Nappo Bappo' in view.groups['Friends'].entities



from click.testing import CliRunner
from contacto.cli import main_cmd
from helpers import fixture, db, mkdb, rmdb, yml_fixture
from os.path import exists
import pytest


@pytest.fixture
def runner():
    mkdb('test')
    return CliRunner()


@pytest.fixture
def tinybin():
    return fixture('tinybin.dat').read_bytes()


def run(runner, arg, set_db=True):
    if set_db:
        arg = f"-o {db} {arg}"
    return runner.invoke(main_cmd, arg)


def test_no_db(runner):
    rmdb()
    result = run(runner, 'get', False)
    assert result.exit_code == 2

    result = run(runner, 'get')
    assert not result.exit_code and exists(db)


def test_get(runner, tinybin):
    result = run(runner, 'get Does/Not/Exist')
    assert not result.exit_code and not result.output

    result = run(runner, 'get Main/TestEntity/Age')
    assert not result.exit_code and result.output == '25\n'

    result = run(runner, 'get //occupation')
    assert not result.exit_code and 'michelin' in result.output

    result = run(runner, 'get -s grp /Albatros/')
    assert not result.exit_code and result.output == 'Friends\n'

    result = run(runner, 'get -s ent -f fam //relationship')
    assert not result.exit_code and 'Dad' in result.output

    result = run(runner, 'get -v 25')
    assert not result.exit_code and 'TestEntity' in result.output

    result = run(runner, 'get -Vv Matthew')
    assert not result.exit_code and 'Logan' in result.output

    result = run(runner, 'get -f /dad/ -Vv logan')
    assert not result.exit_code and result.output.count('Logan') == 1

    result = run(runner, 'get Friends/Albatros/url_bin -r')
    assert not result.exit_code and result.stdout_bytes[:-1] == tinybin


def test_set(runner):
    result = run(runner, 'set -r Family/Sister/Age 18')
    assert not result.exit_code
    result = run(runner, 'get -r Family/Sister/Age')
    assert not result.exit_code and result.output == '18\n'

    result = run(runner, 'set Family/Brother/Age 18')
    assert result.exit_code and result.output.startswith('ERROR')

    catref = 'REF:Family/Dad/catpic'
    result = run(runner, f'set -R Friends/Albatros/url_bin {catref}')
    assert not result.exit_code
    result = run(runner, 'get -f Friends/Albatros/url_bin')
    assert not result.exit_code and result.output.count('BINARY') == 2

    assert run(runner, 'set //url_bin 123').exit_code == 2

    # ref tests
    result = run(runner, 'set Family/Mom/newattr REF:Family/Dad/catpic')
    assert not result.exit_code
    result = run(runner, 'set Family/Mom/catpic REF:Family/Dad/catpic')
    assert result.exit_code and 'REF loop' in result.output



def test_del(runner):
    assert not run(runner, 'del Family/Dad/web').exit_code
    assert not run(runner, 'get Family/Dad/web').output
    assert run(runner, 'del Family/Dad/web').exit_code

    assert not run(runner, 'del Family/Mom/catpic').exit_code
    assert not run(runner, 'del Family/Mom').exit_code
    assert not run(runner, 'del Family').exit_code

    mkdb('test')
    assert not run(runner, 'del Family').exit_code


def test_merge(runner):
    assert not run(runner, 'merge BareEntGrp BareGrp').exit_code
    assert run(runner, 'get BareGrp').output == 'BareEnt\n'

    assert not run(runner, 'merge Family/Mom Family/Dad').exit_code
    result = run(runner, 'get -r Family/Dad/spouse')
    assert not result.exit_code and result.output == 'Family/Dad\n'

    mkdb('test')
    result = run(runner, 'merge Family/Dad Family/Mom')
    assert result.exit_code and 'REF loop' in result.output
    assert not run(runner, 'merge Friends Family').exit_code


def test_import(runner):
    cmd = f"import {yml_fixture('test')}"

    assert not run(runner, cmd).exit_code
    assert not run(runner, 'del Family').exit_code
    assert not run(runner, cmd).exit_code

    rmdb()
    assert not run(runner, cmd).exit_code


def test_export(runner):
    tmp = yml_fixture('temp')
    assert not run(runner, f'export {tmp}').exit_code
    rmdb()
    assert not run(runner, f'import {tmp}').exit_code
    result = run(runner, 'get Family/Dad/age')
    assert not result.exit_code and result.output == '45\n'


def test_plugin_cmd(runner):
    assert not run(runner, 'plugin -l').exit_code
    result = run(runner, 'plugin')
    assert not result.exit_code and result.output.startswith('Plugin summary')



"""The data layer containing all contact data.

Contains objects forming the contact hierarchy: Group, Entity, Attribute; and
the top-most storage container, Storage.

The contact storage is a tree, rooted at the Storage object, and branching
through Groups, Entities and Attributes.
The tree members have parent references towards the root, which wraps
around an SQLite database connection.

Storage manipulation is transactional. For all transformations a safe
variant is provided, `<transform>_safe`, which handles a single transaction.
Unsafe transformations should be wrapped in a transaction manually.
"""
import sqlite3
import pkgutil
from abc import ABC, abstractmethod
from .helpers import DType, bytes_to_attrdata, attrdata_to_bytes, validate_img
from .helpers import Scope, refspec_scope, print_error


DML_SCRIPT = 'resources/dml.sql'


class StorageElement(ABC):
    """An abstract class generalizing Groups, Entities and Attributes.

    These tree elements have scope-unique names and parent references.
    Traversal through the tree is based on names.
    """

    def __init__(self, parent, name):
        """Constructs a tree element from a name and a parent element.
        """
        super().__init__()
        self.parent = parent

        if type(name) is not str or name == '':
            raise Exception("Name must be a non-empty string")
        if '/' in name:
            raise Exception("Illegal character '/' in name")
        self.name = name

    def __str__(self):
        """String representation using scoped name (element refspec).
        """
        return f"{self.parent}/{self.name}"

    @abstractmethod
    def read(self):
        """Loads all element data from the database.
        """
        pass

    @abstractmethod
    def update(self):
        """Saves all element data to the database.

        May modify other (linked) elements.
        """
        pass

    @abstractmethod
    def delete(self):
        """Recursively deletes the subtree rooted at this element.
        """
        pass

    @abstractmethod
    def merge(self, other):
        """Merges another element of the same type into self.

        This deletes the other element and updates self.
        """
        pass

    def get_storage(self):
        """Gets the tree root using upward traversal.

        :return: tree root
        :rtype:  class:`contacto.storage.Storage`
        """
        p = self.parent
        while type(p) is not Storage:
            p = p.parent
        return p

    def get_conn(self):
        """Gets database connection from the root.

        :return: database connection
        :rtype:  class:`sqlite3.Connection`
        """
        return self.get_storage().db_conn

    def update_safe(self):
        """Single-transaction update.

        :return: success
        :rtype:  bool
        """
        try:
            with self.get_conn():
                self.update()
                return True
        except Exception as e:
            print_error(e)
            self.read()
            return False

    def delete_safe(self):
        """Single-transaction delete.

        :return: success
        :rtype:  bool
        """
        try:
            with self.get_conn():
                self.delete()
                return True
        except Exception as e:
            print_error(e)
            # in case of cascade, in-memory data may be corrupt
            self.get_storage().reload()
            return False

    def merge_safe(self, other):
        """Single-transaction merge.

        :return: success
        :rtype:  bool
        """
        try:
            with self.get_conn():
                self.merge(other)
                return True
        except Exception as e:
            print_error(e)
            # in case of cascade, in-memory data may be corrupt
            self.get_storage().reload()
            return False


class Group(StorageElement):
    """An entity group with unique name and pointer to the root storage.

    A Group hosts Entities in a name-indexed dictionary.
    """

    def __init__(self, gid, name, parent):
        """Initialize using DB id, name and parent(Storage)
        """
        super().__init__(parent, name)
        self.id = gid
        self.entities = {}

    def __str__(self):
        return f"{self.name}"

    def create_entity(self, name):
        """Creates a new Entity in this Group.

        Note that Entity names are unique in a Group.

        :param name: Entity name
        :type name:  str
        :return: created Entity
        :rtype:  class:`contacto.storage.Entity`
        """
        cur = self.get_conn().cursor()
        sql = 'INSERT INTO entity VALUES (NULL, ?, NULL, ?)'
        cur.execute(sql, (name, self.id))
        eid = cur.lastrowid

        ent = Entity(eid, name, None, self)
        self.entities[name] = ent
        return ent

    def create_entity_safe(self, name):
        """A single-transaction variant of create_entity.
        """
        try:
            with self.get_conn():
                return self.create_entity(name)
        except Exception as e:
            print_error(e)
            return None

    def read(self):
        """Reads Group from DB.
        """
        cur = self.get_conn().cursor()
        sql = 'SELECT name FROM "group" WHERE id=?'
        cur.execute(sql, [self.id])
        self.name = cur.fetchone()[0]

    def update(self):
        """Saves Group data to DB.
        """
        sql = 'UPDATE "group" SET name=? WHERE id=?'
        self.get_conn().execute(sql, (self.name, self.id))

    def delete(self):
        """Deletes Group (and its Entities) from tree and DB.
        """
        for entity in self.entities.copy().values():
            entity.delete()
        sql = 'DELETE FROM "group" WHERE id=?'
        self.get_conn().execute(sql, [self.id])
        self.parent.groups.pop(self.name, None)

    def merge(self, other):
        """Merges another Group into self.
        """
        for name, oentity in other.entities.copy().items():
            if name in self.entities:
                self.entities[name].merge(oentity)
            else:
                e = self.create_entity(name)
                e.merge(oentity)
        other.delete()


class Entity(StorageElement):
    """The Entity tree element that carries Attributes.

    Entity hosts an Attribute dictionary, its name
    and an optional binary thumbnail (image).
    """

    def __init__(self, eid, name, thumbnail, parent):
        """Initialize using DB data and parent(Group).
        """
        super().__init__(parent, name)
        self.id = eid
        self.thumbnail = thumbnail
        self.attributes = {}
        self.refs = set()

    def create_attribute(self, name, dtype, data):
        """Creates a new Attribute for this Entity.

        Note that Attribute names are unique in an Entity.

        :param name: Attribute name
        :type  name: str
        :param dtype: data type
        :type dtype:  class:`contacto.helpers.DType`
        :return: created thumbnail
        :rtype:  class:`contacto.storage.Attribute`
        """
        cur = self.get_conn().cursor()
        sql = 'INSERT INTO attribute VALUES (NULL, ?, ?, ?, ?)'
        bin_data = attrdata_to_bytes(dtype, data)
        cur.execute(sql, (name, dtype, bin_data, self.id))
        aid = cur.lastrowid

        attr = Attribute(aid, name, dtype, data, self)
        self.attributes[name] = attr

        attr.ref_register()  # blows up if loop is detected
        if name == 'thumbnail':
            self.thumbnail_from_attr()
        return attr

    def create_attribute_safe(self, name, dtype, data):
        """A single-transaction variant of create_attribute.
        """
        try:
            with self.get_conn():
                return self.create_attribute(name, dtype, data)
        except Exception as e:
            print_error(e)
            return None

    def read(self):
        """Reads Entity data from DB.
        """
        cur = self.get_conn().cursor()
        sql = 'SELECT name, thumbnail FROM entity WHERE id=?'
        cur.execute(sql, [self.id])
        self.name, self.thumbnail = cur.fetchone()

    def update(self):
        """Saves Entity data to DB.
        """
        sql = 'UPDATE entity SET name=?, thumbnail=? WHERE id=?'
        self.get_conn().execute(sql, (self.name, self.thumbnail, self.id))

    def delete(self):
        """Deletes Entity (and its Attributes) from DB and tree.
        """
        for attr in self.attributes.copy().values():
            attr.delete()
        sql = 'DELETE FROM entity WHERE id=?'
        self.get_conn().execute(sql, [self.id])
        self.parent.entities.pop(self.name)
        # delete refs pointing to me
        for ref in self.refs.copy():
            ref.delete()

    def merge(self, other):
        """Merges another Entity into self.
        """
        if other.thumbnail and not self.thumbnail:
            self.thumbnail = other.thumbnail
            self.update()
        # redirect all pointers to us
        for attr in other.refs.copy():
            attr.data = self
            attr.update()
        for name, oattr in other.attributes.copy().items():
            if name in self.attributes:
                self.attributes[name].merge(oattr)
            else:
                # create a dummy attribute and merge oattr into it
                attr = self.create_attribute(name, DType.TEXT, '<MERGE>')
                attr.merge(oattr)
        other.delete()

    def thumbnail_from_attr(self):
        """Sets a thumbnail from a "thumbnail" attribute.

        For convenience, a "thumbnail" attribute may carry an entity's
        thumbnail. This is a notification hook to try querying it for data.
        """
        if 'thumbnail' not in self.attributes:
            if self.thumbnail:
                self.thumbnail = None
                self.update()
            return
        thumb = self.attributes['thumbnail']
        ttype, tdata = thumb.get()
        if ttype is DType.BIN and validate_img(tdata):
            if self.thumbnail != tdata:
                self.thumbnail = tdata
                self.update()


class Attribute(StorageElement):
    """The leaf of the storage tree containing contact data.

    An Attribute has a name, data and data type.

    There are 4 types of Attribute data, listed in `contacto.helpers.DType`:
    BIN(ary), TEXT, A(ttribute) X-REF(erence) and E(ntity) X-REF(erence).

    AXREF attributes may form an oriented graph which **must not** form loops.
    Operations Create, Update and Merge are checked for loop induction.

    X-REF attributes register at their targets and are deleted when their
    target is deleted. Hence deleting an Entity or Attribute may cascade.
    These registrations are properly handled on mutations.
    """

    def __init__(self, aid, name, dtype, data, parent):
        """The constructor linking to a parent(Entity).
        """
        super().__init__(parent, name)
        self.__thumb = name == 'thumbnail'
        self.id = aid
        self.type = dtype
        self.data = data
        self.refs = set()

    def ref_register(self):
        """ Registers a reference to its target.
        """
        if self.type.is_xref():
            self.__loop_detect()
            self.data.refs.add(self)

    def ref_unregister(self):
        """ Unregisters a reference from its target.
        """
        if self.type.is_xref():
            self.data.refs.discard(self)

    def __thumb_hook(self):
        """Notifies the Entity for thumbnail update (if applicable).
        """
        if self.__thumb:
            self.parent.thumbnail_from_attr()
        else:
            # propagate thumbnail update recursively
            for ref in self.refs:
                ref.__thumb_hook()

    def read(self):
        """Updates Attribute data from the DB.
        """
        cur = self.get_conn().cursor()
        sql = 'SELECT name, type, data FROM attribute WHERE id=?'
        cur.execute(sql, [self.id])
        self.name, int_type, bin_data = cur.fetchone()
        self.type = DType(int_type)
        self.data = bytes_to_attrdata(self.type, bin_data)
        if self.type.is_xref():
            storage = self.get_storage()
            self.data = storage.elem_from_refid(self.type, self.data)

    def update(self):
        """Saves Attribute data into the DB, checked for loops.
        """
        # check previous data for obsolete XREF registration
        t, d = self.type, self.data
        self.read()  # read original data
        t, self.type = self.type, t  # swap back new data
        d, self.data = self.data, d

        sql = 'UPDATE attribute SET name=?, type=?, data=? WHERE id=?'
        bin_data = attrdata_to_bytes(self.type, self.data)
        self.get_conn().execute(sql, (self.name, self.type, bin_data, self.id))
        # unregister old ref is applicable
        if t.is_xref():
            d.refs.discard(self)
        self.ref_register()
        self.__thumb_hook()

    def delete(self):
        """Deletes Attribute from tree and DB (may cascade!)
        """
        sql = 'DELETE FROM attribute WHERE id=?'
        self.get_conn().execute(sql, [self.id])
        self.parent.attributes.pop(self.name, None)
        # unregister and delete refs pointing to me
        self.ref_unregister()
        for ref in self.refs.copy():
            ref.delete()
        self.__thumb_hook()

    def merge(self, other):
        """Merges another Attribute into self, checked for loops.
        """
        # loop prevention
        self.ref_unregister()
        other.ref_unregister()

        # redirect all pointers to us
        for attr in other.refs.copy():
            attr.data = self
            attr.update()

        self.type = other.type
        self.data = other.data
        self.update()
        other.delete()

    def get(self):
        """Gets the actual data the Attribute references.

        For non-XREF Attributes this returns own data.
        For XREFs, the REF chain is traced to a non-REF source.
        This may be an Entity or a non-XREF Attribute.

        :return: resolved type and data
        :rtype:  (class:`contacto.helpers.DType`, Union[str, bytes])
        """
        if self.type is DType.EXREF:
            return DType.TEXT, str(self.data)
        elif self.type is DType.AXREF:
            return self.data.get()
        return self.type, self.data

    def rotate(self):
        """Rotates the attribute in a logrotate way.

        A single Attribute named "A" is copied into one named "A_1".
        If "A_1" already exists, it is copied into "A_2" and so on.
        This helps maintain an attribute value history.
        """
        self.__rotate(f"{self.name}_1")

    def rotate_safe(self):
        """A single-transaction variant of rotate.
        """
        try:
            with self.get_conn():
                self.rotate()
                return True
        except Exception as e:
            print_error(e)
            return False

    def __rotate(self, nxt_name):
        """Recursive rotate, next Attribute name is passed as accumulator.
        """
        ent = self.parent
        # recursively rotate until end of rotate-chain is found
        if nxt_name in ent.attributes:
            nxt_attr = ent.attributes[nxt_name]
            nxt_name_toks = nxt_name.split('_')
            base = '_'.join(nxt_name_toks[:-1])
            sfx = int(nxt_name_toks[-1]) + 1

            nxt_attr.__rotate(f"{base}_{sfx}")

            nxt_attr.type = self.type
            nxt_attr.data = self.data
            nxt_attr.update()
        else:
            ent.create_attribute(nxt_name, self.type, self.data)

    def __loop_detect(self, stop=None):
        """Traverses reflinks recursively and detects loops.
        """
        if self.type is DType.EXREF:
            return
        elif self.type is DType.AXREF:
            if stop is self:
                raise Exception(f'REF loop detected at {self}')
            elif stop is None:
                stop = self
            self.data.__loop_detect(stop)


class Storage:
    """The root of the storage tree and wrapper of the DB connection.

    Provides communication with an SQLite database and reads the tree from it.
    A Storage object is needed for all of Contacto's functionality.

    Access its Group children with the "groups" name-indexed dictionary.
    """

    def __init__(self, db_file):
        """Initializes the database with a path to the database.

        Optionally, provide a ":memory:" string to create DB in-memory.

        :param db_file: path to the database (or :memory:)
        :type db_file:  str
        """
        self.db_file = db_file
        # connection
        self.db_conn = sqlite3.connect(db_file)
        # executor
        self.db_cur = self.db_conn.cursor()
        self.create_db()
        self.set_foreign_keys(True)

        # load everything from db
        self.reload()

    def __del__(self):
        """Closes the DB connection on deletion.
        """
        self.db_conn.close()

    def set_foreign_keys(self, on):
        """Turns foreign keys ON or OFF.

        :param on: FK mode
        :type on:  bool
        """
        sql = f"PRAGMA foreign_keys = {'ON' if on else 'OFF'}"
        self.db_conn.execute(sql)

    def reload(self):
        """Discards any existing storage tree and reads it anew from DB.

        Constructs the entire tree including Attributes.
        """
        # NAME-indexed group dict
        self.groups = {}
        # ID-indexed helper dicts
        groups_by_id = {}
        entities_by_id = {}
        attributes_by_id = {}

        # attributes to set AXREFs on
        axref_attributes = []

        # load groups
        self.db_cur.execute('SELECT * FROM "group"')
        for gid, name in self.db_cur.fetchall():
            group = Group(gid, name, self)
            groups_by_id[gid] = group
            self.groups[name] = group

        # load entities
        self.db_cur.execute('SELECT * FROM entity')
        for eid, name, thb, gid in self.db_cur.fetchall():
            entity = Entity(eid, name, thb, groups_by_id[gid])
            entities_by_id[eid] = entity
            groups_by_id[gid].entities[name] = entity

        # load attributes
        self.db_cur.execute('SELECT * FROM attribute')
        for aid, name, dtype, data, eid in self.db_cur.fetchall():
            dtype = DType(dtype)
            attr_data = bytes_to_attrdata(dtype, data)
            ent = entities_by_id[eid]

            attribute = Attribute(aid, name, dtype, attr_data, ent)
            attributes_by_id[aid] = attribute
            entities_by_id[eid].attributes[name] = attribute

            if dtype is DType.EXREF:
                attribute.data = entities_by_id[attribute.data]
                attribute.ref_register()
            elif dtype is DType.AXREF:
                axref_attributes.append(attribute)

        # replace AXREFs using actual data
        for attribute in axref_attributes:
            attribute.data = attributes_by_id[attribute.data]
            attribute.ref_register()

    def get_group(self, name):
        """Gets a Group by its name.

        :param name: Group name
        :type name:  str
        :return: found Group or None, if one isn't found
        :rtype:  Union[class:`contacto.storage.Group`, None]
        """
        if name not in self.groups:
            return None
        return self.groups[name]

    def get_entity(self, group_name, name):
        """Gets an Entity by its name and Group name.

        :param group_name: Group name
        :type group_name:  str
        :param name: Entity name
        :type name:  str
        :return: found Entity or None, if one isn't found
        :rtype:  Union[class:`contacto.storage.Entity`, None]
        """
        group = self.get_group(group_name)
        if not group or name not in group.entities:
            return None
        return group.entities[name]

    def get_attribute(self, group_name, entity_name, name):
        """Gets an Attribute by its name and the names of its parents.

        :param group_name: Group name
        :type group_name:  str
        :param entity_name: Entity name
        :type entity_name:  str
        :param name: Attribute name
        :type name:  str
        :return: found Attribute or None, if one isn't found
        :rtype:  Union[class:`contacto.storage.Attribute`, None]
        """
        entity = self.get_entity(group_name, entity_name)
        if not entity or name not in entity.attributes:
            return None
        return entity.attributes[name]

    def create_db(self):
        """Creates DB structure using the prepared DML
        """
        script = pkgutil.get_data(__name__, DML_SCRIPT).decode('utf-8')
        self.db_cur.executescript(script)
        self.db_conn.commit()

    def create_group(self, name):
        """Creates a new Group.
        Mind that Group names are unique in a DB.

        :param name: Group name
        :type name:  str
        :return: created Group
        :rtype:  class:`contacto.storage.Group`
        """
        sql = 'INSERT INTO "group" VALUES (NULL, ?)'
        self.db_cur.execute(sql, [name])
        gid = self.db_cur.lastrowid

        group = Group(gid, name, self)
        self.groups[name] = group
        return group

    def create_group_safe(self, name):
        """A single-transaction variant of the create_group method
        """
        try:
            with self.db_conn:
                return self.create_group(name)
        except Exception as e:
            print_error(e)
            return None

    def get_from_rspec(self, p_rspec):
        """Gets a tree element from a parsed refspec.
        Parsed refspec is a triplet of None|element names scoping the tree.

        :param p_rspec: parsed refspec
        :type  p_rspec: tuple
        :return: tree element if one is found
        :rtype:  Union[None, class:`contacto.storage.StorageElement`]
        """
        scope = refspec_scope(p_rspec)
        g, e, a = p_rspec
        if scope == Scope.GROUP:
            return self.get_group(g)
        if scope == Scope.ENTITY:
            return self.get_entity(g, e)
        if scope == Scope.ATTRIBUTE:
            return self.get_attribute(g, e, a)
        return None

    def elem_from_refid(self, dtype, elem_id):
        """Returns an Entity or Attribute from its database ID.

        The type of element is decided by a DType XREF specifier.
        This method is used to find the target of a XREF from ID only.

        :param dtype: type of element (XREF spec)
        :type dtype:  class:`contacto.helpers.DType`
        :param elem_id: ID of the element
        :type elem_id:  int
        :return: tree element if one is found
        :rtype:  Union[None, class:`contacto.storage.StorageElement`]
        """
        if not dtype.is_xref():
            return None
        if dtype == DType.EXREF:
            sql = 'SELECT g.name, e.name FROM entity as e \
                   LEFT JOIN "group" as g ON (g.id=e.group_id) WHERE e.id=?'
            self.db_cur.execute(sql, [elem_id])
            return self.get_entity(*self.db_cur.fetchone())
        sql = 'SELECT g.name, e.name, a.name FROM attribute as a \
               LEFT JOIN entity as e ON (e.id=a.entity_id) \
               LEFT JOIN "group" as g ON (g.id=e.group_id) WHERE a.id=?'
        self.db_cur.execute(sql, [elem_id])
        return self.get_attribute(*self.db_cur.fetchone())



#! /usr/bin/env python
from setuptools import setup

with open("README.rst", "r") as f:
    long_description = f.read()

setup(
    name='contacto',
    version='0.2.1',
    description='A robust contact manager for social engineering.',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    keywords='contacts, social engineering, Qt, cli',
    author='Adam Zahumensky',
    author_email='zahumada@fit.cvut.cz',
    license='GNU General Public License v3 (GPLv3)',
    url='https://github.com/Dawnflash/contacto',
    packages=['contacto'],
    package_data={'contacto': ['resources/*']},
    entry_points={
        'console_scripts': [
            'contacto = contacto.cli:main',
        ],
        'gui_scripts': [
            'qcontacto = contacto.gui:main',
        ],
    },
    python_requires='>=3.6',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Environment :: Console',
        'Environment :: X11 Applications :: Qt'
    ],
    install_requires=['PyQt5', 'click', 'pyyaml', 'pillow'],
    extras_require={
        'dev': ['sphinx', 'pytest'],
    },
    zip_safe=False,
)



"""Support structures and functions.
"""

from enum import IntEnum
import pkgutil
import importlib
from PIL import Image
from urllib.request import urlopen
import io
import sys
import click


class DType(IntEnum):
    """Attribute data types.
    """
    TEXT = 1,
    BIN = 2,
    AXREF = 3,
    EXREF = 4

    def is_xref(self):
        """True if the data type is a reference.
        """
        return self is self.AXREF or self is self.EXREF


class Scope(IntEnum):
    """Level in the contact hierarchy.
    """
    GROUP = 1,
    ENTITY = 2,
    ATTRIBUTE = 3,
    ATTR_VAL = 4

    @classmethod
    def from_str(cls, s):
        """Creates scope from string.
        """
        if s == 'grp':
            return cls.GROUP
        if s == 'ent':
            return cls.ENTITY
        if s == 'attr':
            return cls.ATTRIBUTE
        raise Exception("Unknown scope")


def bytes_to_attrdata(dtype, bin_data):
    """Parses attribute data from its binary-packed form.

    :param dtype: data type
    :type  dtype: class:`contacto.helpers.DType`
    :param bin_data: data in binary form
    :type  bin_data: bytes
    :return: parsed attribute data
    :rtype:  Union[bytes, str, int]
    """
    if dtype is DType.BIN:
        return bin_data
    if dtype is DType.TEXT:
        return bin_data.decode('utf-8')
    return int.from_bytes(bin_data, byteorder='little')


def attrdata_to_bytes(dtype, attr_data):
    """Packs attribute data into a binary form (for database storage).

    :param dtype: data type
    :type  dtype: class:`contacto.helpers.DType`
    :param attr_data: attribute data
    :type  attr_data: Union[bytes, str, int]
    :return: packed attribute data
    :rtype:  bytes
    """
    if dtype is DType.BIN:
        return attr_data
    if dtype is DType.TEXT:
        return attr_data.encode('utf-8')
    return attr_data.id.to_bytes(4, byteorder='little')


def parse_refspec(rspec):
    """Parses a generic text refspec into its tuple form.

    A generic refspec may have any parts omitted as long as it has at most 3.

    Therefore /E/, G//A, //A, G/E/A are all valid refspecs
    Omitted parts are replaced with `None`

    :param rspec: text refspec
    :type  rspec: str
    :return: parsed refspec
    :rtype:  tuple
    """
    if not rspec:
        return None, None, None
    toks = rspec.split('/')
    ln = len(toks)
    if ln > 3:
        raise Exception('REFSPEC: [GROUP][/[ENTITY][/[ATTRIBUTE]]]')
    toks = [None if tok == '' else tok for tok in toks]
    for _ in range(3 - ln):
        toks.append(None)
    return tuple(toks)


# GROUP/ENTITY/ATTR -> [GROUP, ENTITY, ATTR]
# REFSPEC -> REFERENCE
def parse_ref(rspec):
    """Parse a refspec into a reference and its type (entity or attribute).

    Requires a fully-specified refspec: G/E or G/E/A

    :param rspec: text refspec
    :type  rspec: str
    :raises Exception: the refspec is not a valid entity/attribute reference
    :return: reference type and parsed refspec
    :rtype: (class:`contacto.helpers.DType`, tuple)
    """
    p_rspec = parse_refspec(rspec)
    scope = refspec_scope(p_rspec)
    if scope == Scope.ENTITY:
        return DType.EXREF, p_rspec
    if scope == Scope.ATTRIBUTE:
        return DType.AXREF, p_rspec
    raise Exception('Bad REF signature')


def get_plugins():
    """Uses pkgutil to find plugins among top-level modules.

    :return: found plugins
    :rtype:  dict
    """
    def fam_name(name):
        return name.split('contacto_')[1]

    return {
        fam_name(name): importlib.import_module(name)
        for finder, name, ispkg
        in pkgutil.iter_modules()
        if name.startswith('contacto_')
    }


def run_plugins(storage, whitelist=[]):
    """Runs found and selected plugins.

    :param storage: used storage
    :type  storage: class:`contacto.storage.Storage`
    :param whitelist: allowed plugin names
    :type  whitelist: list, optional
    :return: lists of successful and failed run plugins
    :rtype:  (list, list)
    """
    def wl_compliant(name):
        if len(whitelist) == 0:
            return True
        return name in whitelist

    plugins = get_plugins()
    success, fail = [], []

    for name, plugin in plugins.items():
        if not wl_compliant(name):
            continue
        if hasattr(plugin, 'plugin_init'):
            if plugin.plugin_init(storage):
                success.append(name)
            else:
                fail.append(name)
    return success, fail


def fmatch(needle, haystack):
    """A case-insensitive substring search.
    """
    return needle.casefold() in haystack.casefold()


def dump_lscope(index_filters):
    """Leftmost tree scope to apply when dumping contacts.

    This corresponds to the leftmost directly specified element

    :param index_filters: parsed refspec
    :type  index_filters: tuple
    :return: leftmost scope
    :rtype:  class:`contacto.helpers.Scope`
    """
    scope = Scope.GROUP
    for fil in index_filters:
        if fil:
            scope += 1
        else:
            break
    return Scope(scope)


def refspec_scope(p_rspec):
    """The actual tree (depth) scope a refspec directly represents.

    A not-fully-specified scope such as /E/ represents nothing.

    Fully-specified scopes such as G, G/E and G/E/A do.

    :param p_rspec: parsed refspec
    :type  p_rspec: tuple
    :return: refspec scope
    :rtype:  class:`contacto.helpers.Scope`
    """
    g, e, a = p_rspec
    if g and e and a:
        return Scope.ATTRIBUTE
    if g and e:
        return Scope.ENTITY
    if g and not a:
        return Scope.GROUP
    return None


def attr_val_str(attr, direct):
    """A human-readable representation of attribute data.

    :param attr: contact attribute
    :type  attr: class:`contacto.storage.Attribute`
    :param direct: attribute name won't be printed
    :type  direct: bool, optional
    :return: string representation
    :rtype:  str
    """
    vtype, val = attr.get()
    if direct:
        return val
    s, pfx = '', ''
    if vtype == DType.TEXT:
        s = val
    if vtype == DType.BIN:
        s = f"<BINARY, {size_str(val)}>"
    if attr.type.is_xref():
        pfx = f'[-> {attr.data}] '
    if attr.type == DType.EXREF and not direct:
        s = ''  # no need to print the spec twice
    return f"{pfx}{s}"


def size_str(blob):
    """Gets binary data size in human-readable units.

    :param blob: binary data
    :type  blob: bytes
    :return: data size with units
    :rtype:  str
    """
    size = len(blob)
    sfx = 'B'
    if size > 1024:
        size /= 1024
        sfx = 'kB'
    if size > 1024:
        size /= 1024
        sfx = 'MB'
    return f"{int(size)}{sfx}"


def validate_img(data):
    """Checks if data are an image.

    :param data: binary data
    :type  blob: bytes
    :return: true if data is an image
    :rtype:  bool
    """
    try:
        bio = io.BytesIO(data)
        img = Image.open(bio)
        img.verify()
    except Exception:
        return False
    return True


def parse_valspec(value):
    """Parses an attribute value specifier (used in data input/import).

    :param value: value specifier
    :type  value: str
    :return: parsed attr value with its type
    :rtype:  (class:`contacto.helpers.DType`, Union[bytes, str, tuple])
    """
    if not value:
        return None, None
    if value.startswith('FILE:'):
        fname = value.split('FILE:')[1]
        with open(fname, 'rb') as f:
            return DType.BIN, f.read()
    if value.startswith('URL:'):
        url = value.split('URL:')[1]
        with urlopen(url) as f:
            return DType.BIN, f.read()
    if value.startswith('REF:'):
        # return parsed reference
        return parse_ref(value.split('REF:')[1])
    return DType.TEXT, value


def print_error(err):
    """Prints a formatted error message to stderr.
    """
    click.echo("{}: {}".format(
        click.style("ERROR", fg='red', bold=True),
        err, file=sys.stderr
    ))


def print_warning(warn):
    """Prints a formatted warning message to stderr.
    """
    click.echo("{}: {}".format(
        click.style("WARN", fg='yellow', bold=True),
        warn, file=sys.stderr
    ))



import os
import re
import pathlib
import atexit
from contacto.serial import Serial
from contacto.storage import Storage


def fixture(name=None):
    path = pathlib.Path(__file__).parent / 'fixtures'
    if name:
        path = path / name
    return path


def yml_fixture(name):
    return fixture(f"_{name}.yml")


def mkyaml():
    for yml in yamls:
        if not fixture(f"{yml}.yml").exists():
            continue
        ydata = fixture(f"{yml}.yml").read_text()
        ydata = ydata.replace('<FIXTURES_PATH>', str(fixture().absolute()))
        yml_fixture(yml).write_text(ydata)


def mkdb(name):
    rmdb()
    stor = Storage(str(db))
    ser = Serial(stor)
    with open(yml_fixture(name), 'r') as f:
        ser.import_yaml(f)


def rmdb():
    path = fixture(db)
    if os.path.exists(path):
        path.unlink()


def mkstor():
    return Storage(':memory:')


def cleanup():
    for yml in yamls:
        path = yml_fixture(yml)
        if os.path.exists(path):
            path.unlink()
    rmdb()


db = fixture('data.db')
yamls = [
    'test',
    'temp'
]

mkyaml()
atexit.register(cleanup)
